
package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.border.*;

import com.skynamics.lib.swing.NumEdit;
import com.skynamics.lib.swing.SimpleGridBag;

public class EdgeEditor extends JPanel 
implements ActionListener, FocusListener
{
	private Edge myEdge;
	private Journey myJourney;
	private JTextField roadName; 
	private NumEdit length;
	private JComboBox from;
	private JComboBox to;
	private JComboBox fromBranch;
	private JComboBox toBranch;
	private ListDataListener ldl;
	private Vertex[] vertex;
	private JLabel[] tflabel = new JLabel[6];
	private double[] matrix = new double[6];
	private NumberFormat nf = NumberFormat.getInstance ();
	private JLabel fromToLabel;
	private JLabel mapNameLabel;
	private JLabel mapSizeLabel;
	private JLabel mapId;
	// private JTextField journeyId;
	private JTextField journeyDesc;
	
	public EdgeEditor (Edge edge, Journey j, Vertex[] v)
	{
		setLayout (new SimpleGridBag ());
		myEdge = edge;
		myJourney = j;
		// assign edge from/to to journey unless journey already matches
		if ((j.from != edge.vertex[1]) || (j.to != edge.vertex[0]))
		{
			j.from = edge.vertex[0]; 
			j.to = edge.vertex[1];
		}

		vertex = v;
		JPanel tmpBox;
		JButton tmpButton;
			
		JPanel edgePanel = new JPanel ();
		add (edgePanel, "x=0 y=1 f=h");
		edgePanel.setBorder (BorderFactory.createTitledBorder ("Edge"));
		edgePanel.setLayout (new SimpleGridBag ());

		edgePanel.add (new JLabel ("Road:"), "a=w x=0 y=0 wy=.2 ir=5 wx=.3");
		edgePanel.add (roadName = new JTextField (), "f=h wx=.7");
		roadName.addFocusListener (this);
		edgePanel.add (new JLabel ("Length:"), "nl a=w wy=.2 ir=5");
		edgePanel.add (new NumEdit.AddSpinPanel (length = new NumEdit ()), "f=h");
		length.addFocusListener (this);
		JPanel flagPanel = new JPanel ();
		JCheckBox flagCoastal, flagSingleTrack, flagMountain, flagDualCW, flagPrivate;
		flagPanel.setLayout (new SimpleGridBag ());
		flagPanel.setBorder (BorderFactory.createTitledBorder ("Flags"));
		flagPanel.add (flagCoastal = new JCheckBox ("Coastal"), "il=3 a=w wx=.5");
		flagPanel.add (flagSingleTrack = new JCheckBox ("Single Trk"), "a=w wx=.5 ir=3");
		flagPanel.add (flagMountain = new JCheckBox ("Mountain"), "nl a=w il=3");
		flagPanel.add (flagDualCW = new JCheckBox ("Dual C.W."), "a=w ir=3");
		flagPanel.add (flagPrivate = new JCheckBox ("Private"), "nl a=w il=3");
		/* flagPanel.add (new JCheckBox ("Or Other"), "a=w ir=3"); */
		edgePanel.add (flagPanel, "nl w=2 wy=.6 f=b");
		
		JPanel fromPanel = new JPanel ();
		fromPanel.setLayout (new SimpleGridBag ());
		fromPanel.setBorder (BorderFactory.createTitledBorder ("Connecting..."));
		fromPanel.add (from = new JComboBox (new VertexComboBoxModel ()), "wx=.5 f=h i=5");
		fromPanel.add (new JLabel ("->"));
		fromPanel.add (fromBranch = new JComboBox (new BranchComboBoxModel (from)), "wx=.5 f=h i=5");
		from.addActionListener (this);
		fromBranch.addActionListener (this);

		JPanel toPanel = new JPanel ();
		toPanel.setLayout (new SimpleGridBag ());
		toPanel.setBorder (BorderFactory.createTitledBorder ("... and ..."));
		toPanel.add (to = new JComboBox (new VertexComboBoxModel ()), "wx=.5 f=h i=5");
		toPanel.add (new JLabel ("->"));
		toPanel.add (toBranch = new JComboBox (new BranchComboBoxModel (to)), "wx=.5 f=h i=5");
		to.addActionListener (this);
		toBranch.addActionListener (this);

		JButton viaButton = new JButton ("Edit Vias...");
	
		edgePanel.add (fromPanel, "il=10 wx=1 f=h nl w=2");
		edgePanel.add (toPanel, "nl il=10 ib=5 wx=1 f=h w=2");
		edgePanel.add (viaButton, "nl a=e w=2");
		
		add (edgePanel, "x=0 y=0 f=h wx=1 i=5");

		JPanel journeyPanel = new JPanel ();
		journeyPanel.setBorder (BorderFactory.createTitledBorder ("Journey"));
		journeyPanel.setLayout (new SimpleGridBag ());
		journeyPanel.add (new JLabel ("Journey Id:"), "a=w il=5 ir=5");
		// journeyPanel.add (journeyId = new JTextField (), "w=2 f=h ir=5");
		// journeyId.addFocusListener (this);
		journeyPanel.add (new JLabel ("Description:"), "nl a=w ir=5 il=5");
		journeyPanel.add (journeyDesc = new JTextField (), "w=2 f=h wx=1 ir=5");
		journeyDesc.addFocusListener (this);
		journeyPanel.add (new JLabel ("Direction:"), "nl a=w ir=5 il=5 wx=.1");
		journeyPanel.add (fromToLabel = new JLabel ("(no start and destination selected)"), " il=5  f=h wx=.8 a=w");
		updateJourney ();
		JButton fromToButton;
		journeyPanel.add (fromToButton = new JButton ("Swap"), "nl a=e ir=5 wx=.8");
		fromToButton.addActionListener (new ActionListener () {
			public void actionPerformed (ActionEvent e) {
				Vertex tmp = myJourney.from;
				myJourney.from = myJourney.to;
				myJourney.to = tmp;
				updateJourney ();
			}
		});

		add (journeyPanel, "nl f=h wx=1 il=5 ir=5");

		JPanel mapPanel = new JPanel ();
		mapPanel.setBorder (BorderFactory.createTitledBorder ("Map"));
		mapPanel.setLayout (new SimpleGridBag ());
		mapPanel.add (new JLabel ("Map Id:"), "il=5 a=w f=h wx=.2");
		mapPanel.add (mapId = new JLabel (""), "f=h wx=.8 ir=5");
		mapPanel.add (new JLabel ("File:"), "il=5 nl a=w it=3 ib=3");
		mapPanel.add (mapNameLabel = new JLabel (""), "f=h it=3 ib=3");
		mapPanel.add (new JLabel ("Size:"), "il=5 nl a=w it=3 ib=3");
		mapPanel.add (mapSizeLabel = new JLabel (""), "f=h it=3 ib=3");
		mapPanel.add (new JLabel ("Transform:"), "nl a=nw il=5");

		mapPanel.add (new JLabel ("FIXME"), "f=h ir=5");

		add (mapPanel, "nl f=h wx=1 i=5");

		updateMap ();
	}

	public void updateEdge ()
	{
		roadName.setText (myEdge.road);
		length.setValue  (myEdge.length);
		from.setSelectedItem (myEdge.vertex[0].id);
		fromBranch.setSelectedItem (myEdge.connector[0]);
		to.setSelectedItem (myEdge.vertex[1].id);
		toBranch.setSelectedItem (myEdge.connector[1]);
		// FIXME update flags
	}
		
	public void focusLost (FocusEvent e)
	{
		ActionEvent ae = new ActionEvent (e.getSource (), 0, null);
		actionPerformed (ae);
	}
	public void focusGained (FocusEvent e) { /* ignore */ }
		
	public void actionPerformed (ActionEvent e)
	{
		if ((e.getSource () == from) && (from.getSelectedIndex () >= 0))
		{
			Vertex selected = vertex[from.getSelectedIndex ()];
			if (myJourney.to == myEdge.vertex[0])
			{
				myJourney.to = selected;
			}
			else
			{
				myJourney.from = selected;
			}
			myEdge.vertex[0] = selected;
			updateJourney ();
		}
		else if ((e.getSource () == to) && (to.getSelectedIndex () >= 0))
		{
			Vertex selected = vertex[to.getSelectedIndex ()];
			if (myJourney.from == myEdge.vertex[1])
			{
				myJourney.from = selected;
			}
			else
			{
				myJourney.to = selected;
			}
			myEdge.vertex[1] = selected;
			updateJourney ();
		}
		else if (e.getSource () == fromBranch)
		{
			myEdge.connector[0] = fromBranch.getSelectedItem ().toString ();
		} 
		else if (e.getSource () == toBranch)
		{
			myEdge.connector[1] = toBranch.getSelectedItem ().toString ();
		} 
/*
		else if (e.getSource () == journeyId)
		{
			myJourney.id = journeyId.getText ();
		}
*/
		else if (e.getSource () == journeyDesc)
		{
			//myJourney.description = journeyDesc.getText ();
		}
		else if (e.getSource () == length)
		{
			myEdge.length = (int) length.getValue ();
		}
		else if (e.getSource () == roadName)
		{
			myEdge.road = roadName.getText ();
		}
		
	}

	public void setVertexArray (Vertex[] v)
	{
		vertex = v;
		fireChange ();
	}

	public void updateMap ()
	{
		
		if (myEdge.map == null) 
		{
			mapNameLabel.setText ("");
			mapSizeLabel.setText ("");
			mapId.setText ("");
		}
		else
		{
			mapNameLabel.setText (myEdge.map.filename);
			mapId.setText (myEdge.map.id);
			if (myEdge.map.size == null)
			{
				mapSizeLabel.setText ("");
			}
			else
			{
				mapSizeLabel.setText (myEdge.map.size.width + " x " + myEdge.map.size.height);
			}
		}
		mapNameLabel.invalidate ();
		mapSizeLabel.invalidate ();
	}

	public void updateJourney ()
	{
		if ((myJourney.from == null) || (myJourney.to == null)) 
		{
			return;
		}
		fromToLabel.setText ("From " + myJourney.from.name + " to " + myJourney.to.name);
		fromToLabel.repaint ();
		// journeyId.setText (myJourney.id);
		// journeyDesc.setText (myJourney.description);
	}

	public void setEdge (Edge edge)
	{
		myEdge = edge;
		updateMap ();
		updateEdge ();
	}

	public void setJourney (Journey j)
	{
		myJourney = j;
		updateJourney ();
	}
		
	private void fireChange ()
	{
		Object[] listeners = listenerList.getListenerList();
		for (int i = listeners.length-2; i>=0; i-=2) {
			if (listeners[i]==ListDataListener.class) 
			{
				((ListDataListener)listeners[i+1]).contentsChanged (new ListDataEvent (this, ListDataEvent.CONTENTS_CHANGED, 0, vertex.length));
			}
         }
	}

	/**
	 * Selection slaved to selection in "master" combo box.
	 */
	private class BranchComboBoxModel implements ComboBoxModel, ActionListener
	{
		private Object selection;
		private EventListenerList listenerList = new EventListenerList ();
		private int selInd = 0;

		public BranchComboBoxModel (JComboBox master)
		{
			master.addActionListener (this);
		}
		public void actionPerformed (ActionEvent e)
		{
			Object[] listeners = listenerList.getListenerList();
			selInd = ((JComboBox) e.getSource ()).getSelectedIndex ();
			for (int i = listeners.length-2; i>=0; i-=2) {
				if (listeners[i]==ListDataListener.class) 
				{
					((ListDataListener)listeners[i+1]).contentsChanged (new ListDataEvent (this, ListDataEvent.CONTENTS_CHANGED, 0, 0));
				}
			}
		}
		public Object getSelectedItem () { return selection; }
		public void setSelectedItem (Object o) { selection = o; };
		public void addListDataListener (ListDataListener l) 
			{ listenerList.add (ListDataListener.class, l); }
		public void removeListDataListener (ListDataListener l) 
			{ listenerList.remove (ListDataListener.class, l); }
		public int getSize () { return (((vertex == null)||(selInd<0)||(selInd>=vertex.length))? 0 : vertex[selInd].connector.length); };
		public Object getElementAt (int i) { return (vertex[selInd].connector[i]); };
	};

	private class VertexComboBoxModel implements ComboBoxModel
	{
		private Object selection;
		public Object getSelectedItem () { return selection; }
		public void setSelectedItem (Object o) { selection = o; };
		public void addListDataListener (ListDataListener l) 
			{ listenerList.add (ListDataListener.class, l); }
		public void removeListDataListener (ListDataListener l) 
			{ listenerList.remove (ListDataListener.class, l); }
		public int getSize () { return ((vertex == null)? 0 : vertex.length); };
		public Object getElementAt (int i) { return (vertex[i].id); };
	};

}
