
package org.remote.roadhog;

import java.awt.*;
import java.io.*;
import java.util.*;

import org.w3c.dom.*;
import magick.*;

public class Illustration extends XMLObject
{
	public String filename;
	public Dimension size;
	private double rotation = 0;
	private Rectangle crop = null;
	private HashMap text = new HashMap ();

	/**
	 * Creates new Photo based on XML node
	 */
	public Illustration (Node node)
	throws ParseException
	{
		NodeList sublist;
		Element ill = (Element) node;

		sublist = ill.getElementsByTagName ("filename");
		filename = getTextChild (sublist.item(0));

		sublist = ill.getElementsByTagName ("width");
		int w = Integer.parseInt (getTextChild (sublist.item(0)));

		sublist = ill.getElementsByTagName ("height");
		int h = Integer.parseInt (getTextChild (sublist.item(0)));

		size = new Dimension (w, h);

		sublist = ill.getElementsByTagName ("crop");
		if (sublist.getLength () != 0)
		{
			crop = parseRect ((Element) sublist.item (0));
		}

		sublist = ill.getElementsByTagName ("rotate");
		if (sublist.getLength () != 0) rotation = Double.parseDouble (getTextChild (sublist.item(0)));
		
		parseAttributeWithLanguage (ill, "text", text);
	}

	/**
	 * writes XML serialized version of photo to stream
	 */
	public void toXML (PrintStream o, String indent)
	{
		o.println (indent + "<illustration>");
		o.println (indent + "   <filename>" + stringToXML (filename) + "</filename>");
		o.println (indent + "   <width>" + size.width + "</width>");
		o.println (indent + "   <height>" + size.height + "</height>");
		if (rotation != 0) o.println (indent + "   <rotate>" + rotation + "</rotate>");
		if (crop != null) o.println (indent + "   <crop><x>"+crop.x+"</x><y>"+crop.y+"</y><width>"+crop.width+"</width><height>"+crop.height+"</height></crop>");
		writeAttributeWithLanguage (o, indent + "   ", "text", text);
		o.println (indent + "</illustration>");
	}

	public String getText (String language)
	{
		return getAttributeInLanguage (text, language);
	}

	public void saveAs (String filename)
	throws magick.MagickException, java.io.IOException
	{
		ImageInfo ilii = new ImageInfo (this.filename);
		MagickImage tmp1 = new MagickImage (ilii);

		if (rotation != 0)
		{
			MagickImage tmp2 = tmp1.rotateImage (rotation);
			tmp1.destroyImage ();
			tmp1 = tmp2;
		}

		if (crop != null)
		{
			MagickImage tmp2 = tmp1.cropImage (crop);
			tmp1.destroyImage ();
			tmp1 = tmp2;
		}

		MagickImage scaled = tmp1.scaleImage (size.width, size.height);
		tmp1.destroyImage ();
		scaled.setFileName (filename);
		scaled.writeImage (ilii);
		scaled.destroyImage ();
	}


}
