
package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;

import org.w3c.dom.*;

import org.apache.xerces.parsers.DOMParser;

class Map extends XMLObject
{
	public String id = "";
	public String filename = "";
	public String draftfilename = null;
	public Dimension size = new Dimension ();
	public int[] rowDivider;
	public int[] colDivider; 
	public HashMap mapInfo;
	public int lastUsedCorrectionPercent = 100;
	// public int scale = 1;

	private static final int PROJ_OSGS = 1;
	private static final int PROJ_DIRECT = 0;

	private int projection = 0;
	private AffineTransform transform;
	private Hashtable corrections;
	private Hashtable deletions;
	private Vector divisions;
	private Point2D.Double helperPoint = new Point2D.Double ();

	// used in computeCorrection, defined here for speed
	private final Rectangle testRect = new Rectangle ();
	private final Point[] neighbours = new Point[2];
	private final double distance[] = new double[2];
	private boolean squareOnly = false;

	/**	
	 * Creates an empty Map object.
	 */
	public Map ()
	{
		/* empty */
	}

	/**
	 * Creates a new Map object from the given XML node.
	 */
	public Map (Node node)
	throws ParseException
	{
        NodeList sublist;
        Element map = (Element) node;

        sublist = map.getElementsByTagName ("id");
        id = getTextChild (sublist.item(0));

        sublist = map.getElementsByTagName ("filename");
        filename = getTextChild (sublist.item(0));
        sublist = map.getElementsByTagName ("draft-filename");
        if (sublist.getLength () == 1) draftfilename = getTextChild (sublist.item(0));

        sublist = map.getElementsByTagName ("square-only");
        if (sublist.getLength () == 1) squareOnly = (Integer.parseInt (getTextChild (sublist.item(0))) == 1);

        sublist = map.getElementsByTagName ("width");
        int width = Integer.parseInt (getTextChild (sublist.item(0)));
        sublist = map.getElementsByTagName ("height");
        int height = Integer.parseInt (getTextChild (sublist.item(0)));
        // sublist = map.getElementsByTagName ("scale");
        // if (sublist.getLength () == 1) scale = Integer.parseInt (getTextChild (sublist.item(0)));
	
        sublist = map.getElementsByTagName ("rowdivide");
		int rdcount = sublist.getLength ();
		if (rdcount > 0)
		{
			rowDivider = new int[rdcount];
			for (int i=0; i<rdcount; i++)
        		rowDivider[i] = Integer.parseInt (getTextChild (sublist.item(i)));
		}
        sublist = map.getElementsByTagName ("coldivide");
		int cdcount = sublist.getLength ();
		if (cdcount > 0)
		{
			colDivider = new int[cdcount];
			for (int i=0; i<cdcount; i++)
        		colDivider[i] = Integer.parseInt (getTextChild (sublist.item(i)));
		}

		/*
		 * If no explicit map division settings, split map into equal width/height rows/cols
		 */
		if ((colDivider == null) && (rowDivider == null))
		{
			int squareWidth = 0;
			int squareHeight = 0;
			sublist = map.getElementsByTagName ("square-width");
			if (sublist.getLength () > 0) 
			{
				squareWidth = Integer.parseInt (getTextChild (sublist.item (0)));
			}
			sublist = map.getElementsByTagName ("square-height");
			if (sublist.getLength () > 0) 
			{
				squareHeight = Integer.parseInt (getTextChild (sublist.item (0)));
			}
			if (squareWidth > 0)
			{
				if (squareOnly) width -= (width % squareWidth);
				int cdc = width / squareWidth - 1;
				if (width % squareWidth > 0) cdc++;
				colDivider = new int[cdc];
				for (int i=0; i<cdc; i++) colDivider[i] = (i+1) * squareWidth;
			}
			if (squareHeight > 0)
			{
				if (squareOnly) height -= (height % squareHeight);
				int rdc = height / squareHeight - 1;
				if (height % squareHeight > 0) rdc++;
				rowDivider = new int[rdc];
				for (int i=0; i<rdc; i++) rowDivider[i] = (i+1) * squareHeight;
			}
		}

		size = new Dimension (width, height);

        sublist = map.getElementsByTagName ("projection");
		if (sublist.getLength () > 0)
		{
			String pname = getTextChild (sublist.item (0));
			if ("OSGS".equals (pname))
			{
				projection = PROJ_OSGS;
				transform = null;
			}
			else if ("direct".equals (pname))
			{
				projection = PROJ_DIRECT;
			}
			else
			{
				throw new ParseException ("Invalid projection: '" + pname + "'");
			}
		}

		if ((projection == PROJ_DIRECT) || (projection == PROJ_OSGS))
		{
        	sublist = map.getElementsByTagName ("transform");
			if (sublist.getLength () == 0)
			{
				throw new ParseException ("projection requires transform element");
			}
        	StringTokenizer tok = new StringTokenizer (getTextChild (sublist.item(0)), ";");
			double[] dbl = new double[6];
			try
			{
				for (int i=0; i<6; i++)
					dbl[i] = Double.parseDouble (tok.nextToken ());
			}
			catch (Exception ex)
			{
				throw new ParseException ("Wrong format in 'transform' tag");
			}
			transform = new AffineTransform (dbl);
		}

        sublist = map.getElementsByTagName ("correction-group");
		int corcount = sublist.getLength ();
		if (corcount > 0)
		{
			for (int i=0; i<corcount; i++)
			{
				parseCorrectionGroup ((Element) sublist.item(i));
			}
		}

	}

	private void parseCorrectionGroup (Element e)
	{
		Point calib;
		Point percent;
		NodeList sublist;
		sublist = e.getElementsByTagName ("correction-offset");
		if (sublist.getLength () > 0)
		{
			calib = parsePoint ((Element)sublist.item (0));
		}
		else
		{
			calib = new Point (0,0);
		}
		sublist = e.getElementsByTagName ("correction-percent");
		if (sublist.getLength () > 0)
		{
			percent = parsePoint ((Element)sublist.item (0));
			lastUsedCorrectionPercent = percent.x;
		}
		else
		{
			percent = new Point (100,100);
		}
			
		sublist = e.getElementsByTagName ("correction");
		int corcount = sublist.getLength ();
		if (corcount > 0)
		{
			for (int i=0; i<corcount; i++)
			{
				parseCorrection ((Element)sublist.item (i), calib, percent);
			}
		}
			
		sublist = e.getElementsByTagName ("deletion");
		int delcount = sublist.getLength ();
		if (delcount > 0)
		{
			for (int i=0; i<delcount; i++)
			{
				parseDeletion ((Element)sublist.item (i), calib, percent);
			}
		}

		sublist = e.getElementsByTagName ("division");
		int divcount = sublist.getLength ();
		if (divcount > 0)
		{
			for (int i=0; i<divcount; i++)
			{
				parseDivision ((Element)sublist.item (i), calib, percent);
			}
		}
	}

	// TODO error handling
	private void parseCorrection (Element n, Point cal, Point perc)
	{
		NodeList sublist;
		sublist = n.getElementsByTagName ("from");
		if (sublist.getLength () == 1)
		{
			Point f = parsePoint ((Element)sublist.item (0));
			sublist = n.getElementsByTagName ("to");
			if ((f != null) && (sublist.getLength () == 1))
			{
				Point t = parsePoint ((Element)sublist.item (0));
				if (t != null)
				{
					f.x = f.x * perc.x / 100 + cal.x;
					f.y = f.y * perc.y / 100 + cal.y;
					t.x = t.x * perc.x / 100 + cal.x;
					t.y = t.y * perc.y / 100 + cal.y;
					addCorrection (f, t);
				}
			}
		}
	}

	private void parseDeletion (Element n, Point cal, Point perc)
	{
		NodeList sublist;
		Point f = parsePoint (n);
		f.x = f.x * perc.x / 100 + cal.x;
		f.y = f.y * perc.y / 100 + cal.y;
		addDeletion (f);
	}

	private void parseDivision (Element n, Point cal, Point perc)
	{
		NodeList sublist;
		sublist = n.getElementsByTagName ("from");
		if (sublist.getLength () == 1)
		{
			Point f = parsePoint ((Element)sublist.item (0));
			sublist = n.getElementsByTagName ("to");
			if ((f != null) && (sublist.getLength () == 1))
			{
				Point t = parsePoint ((Element)sublist.item (0));
				if (t != null)
				{
					f.x = f.x * perc.x / 100 + cal.x;
					f.y = f.y * perc.y / 100 + cal.y;
					t.x = t.x * perc.x / 100 + cal.x;
					t.y = t.y * perc.y / 100 + cal.y;
					addDivision (f, t);
				}
			}
		}
	}


	public void addCorrection (Point src, Point dst)
	{
		if (corrections == null) corrections = new Hashtable ();
		corrections.put (src, dst);
	}
	public void addDivision (Line div)
	{
		if (divisions == null) divisions = new Vector ();
		divisions.addElement (div);
	}
	public void addDivision (Point src, Point dst)
	{
		Line l = new Line ();
		l.p1 = src; l.p2 = dst;
		addDivision (l);
	}
	public void addDeletion (Point del)
	{
		if (deletions == null) deletions = new Hashtable ();
		deletions.put (del, null);
	}

	public void clearCorrections ()
	{
		corrections = null;
	}
	public void clearDeletions ()
	{
		deletions = null;
	}
	public void clearDivisions ()
	{
		divisions = null;
	}
	public Vector getDivisions ()
	{
		return divisions;
	}

	public void toXML (PrintStream o, String indent)
	{
		o.println (indent + "<map>");
		o.println (indent + "   <id>" + id + "</id>");
		o.println (indent + "   <filename>" + filename + "</filename>");
		if (draftfilename != null) o.println (indent + "   <draft-filename>" + draftfilename + "</draft-filename>");
		o.println (indent + "   <width>" + size.width + "</width>");
		o.println (indent + "   <height>" + size.height + "</height>");
		// if (scale != 1) o.println (indent + "   <scale>" + scale + "</scale>");

		if (projection == PROJ_OSGS)
		{
			o.println (indent + "   <projection>OSGS</projection>");
		}	
		else
		{
			o.println (indent + "   <projection>DIRECT</projection>");
		}	
		if ((projection == PROJ_OSGS) || (projection == PROJ_DIRECT))
		{
			o.print (indent + "   <transform>");
			double[] dbl = new double[6];
			transform.getMatrix (dbl);
			for (int i=0; i<6; i++) 
			{
				if (i>0) o.print(";");
				o.print(dbl[i]); 
			}
			o.println("</transform>");
		}
		
		/*
		 * If row and col dividers make a pattern of equal squares, save it as
		 * square-height and square-width instead of individual rows/colums
		 */
		int sqh = 0;
		int sqw = 0;
		if (rowDivider != null) 
		{
			sqh = rowDivider[0];
			for (int i = 1; i<rowDivider.length; i++) 
				if (rowDivider[i] != (i+1) * sqh)
				{
					sqh = 0;
					break;
				}
		}
		if (colDivider != null) 
		{
			sqw = colDivider[0];
			for (int i = 1; i<colDivider.length; i++) 
				if (colDivider[i] != (i+1) * sqw)
				{
					sqw = 0;
					break;
				}
		}

		if ((sqw != 0) && (sqh != 0))
		{
			o.println (indent + "   <square-width>" + sqw + "</square-width>");
			o.println (indent + "   <square-height>" + sqh + "</square-height>");
		}
		else
		{
			if (rowDivider != null) for (int i = 0; i<rowDivider.length; i++) 
					o.println (indent + "   <rowdivide>" + rowDivider[i] + "</rowdivide>");
			if (colDivider != null) for (int j = 0; j<colDivider.length; j++) 
				o.println (indent + "   <coldivide>" + colDivider[j] + "</coldivide>");
		}

		if (corrections != null)
		{
			o.println (indent + "   <correction-group>");
			for (Enumeration e = corrections.keys (); e.hasMoreElements (); )
			{
				Point s = (Point) e.nextElement ();
				Point d = (Point) corrections.get (s);
				o.println (indent + "      <correction><from><x>"+s.x+"</x><y>"+s.y+"</y></from><to><x>"+d.x+"</x><y>"+d.y+"</y></to></correction>");
			}
			if (deletions != null)
			{
				for (Enumeration e = deletions.keys (); e.hasMoreElements (); )
				{
					Point s = (Point) e.nextElement ();
					o.println (indent + "      <deletion><x>"+s.x+"</x><y>"+s.y+"</y></deletion>");
				}
			}
			if (divisions != null)
			{
				for (Enumeration e = divisions.elements (); e.hasMoreElements (); )
				{
					Line l = (Line) e.nextElement ();
					o.println (indent + "      <division><from><x>"+l.p1.x+"</x><y>"+l.p1.y+"</y></from><to><x>"+l.p2.x+"</x><y>"+l.p2.y+"</y></to></division>");
				}
			}

			o.println (indent + "   </correction-group>");
		}
				
		o.println (indent + "</map>");
	}

	public Hashtable getCorrections ()
	{
		return corrections;
	}	

	public String getHandle ()
	{
		StringBuffer handle;
		handle = new StringBuffer (id);
		for (int i=0; i<handle.length (); i++)
		{
			if ((handle.charAt (i) == ' ') || (handle.charAt (i) == '/')) handle.deleteCharAt (i--);
		}
		return (handle.toString ().toLowerCase ());
	}

	public void mapGeoToPixel (Point2D from, Point2D to)
	{
		if ((transform != null) && (projection == PROJ_DIRECT))
		{
			transform.transform (from, to);
		}
		else if ((transform != null) && (projection == PROJ_OSGS))
		{
			// first compute the British Grid easting and northing from the input
			double[] grid = GridConverter.degreeToGrid (from.getY (), from.getX ());
			helperPoint.x = grid[0]; helperPoint.y = grid[1];
			transform.transform (helperPoint, to);
		}
		else
		{
			to.setLocation (from);
		}
	}

	public AffineTransform getTransform ()
	{
		return transform;
	}

	public boolean isPixelOnMap (Point p)
	{
		return ((p.x >= 0) && (p.y >= 0) && (p.x < size.width) && (p.y < size.height));
	}

	public boolean isPixelDeleted (Point pixel)
	{
		if (deletions == null) return false;
		return deletions.containsKey (pixel);
	}

	public double[] computeCorrection (Point thisPoint)
	{

		if (corrections == null)
		{
			return new double[] { 0, 0 };
		}

		Point test = (Point) corrections.get (thisPoint);
		if (test != null)
		{
			return new double[] { test.x - thisPoint.x, test.y - thisPoint.y };
		}

		int found = 0;
		boolean doWeedout = false;
		int weedout = -1;

		for (int radius = 25; radius < 90; radius += 20)
		{
			testRect.setBounds (thisPoint.x - radius, thisPoint.y - radius, 2 * radius, 2 * radius);
			found = 0;
			double bestDistance = 0;
			
			for (Enumeration e = corrections.keys (); e.hasMoreElements (); )
			{
				Point t = (Point) e.nextElement ();
				if (testRect.contains (t))
				{

					// never use a neighbour if it's on the other side of a division
					if (divisions != null)
					{
						boolean stop = false;
						for (int i=0; i<divisions.size (); i++)
						{
							Line div = (Line) divisions.elementAt (i);
							if (Line2D.linesIntersect (div.p1.x, div.p1.y, div.p2.x, div.p2.y,
								t.x, t.y, thisPoint.x, thisPoint.y))
							{
								stop = true;
								break;
							}
						}
						if (stop) continue;
					}

					doWeedout = false;

					// if we haven't found enough neighbours yet, use this in any case
					if (found < neighbours.length)
					{
						distance[found] = t.distance (thisPoint);
						bestDistance = Math.max (distance[found], bestDistance);
						neighbours[found++] = t;
						doWeedout = true;
					}
					// else use it only if it is nearer than one of the others
					else
					{
						double thisDistance = t.distance (thisPoint);
						if (thisDistance < bestDistance)
						{
							for (int i=0; i<neighbours.length; i++)
							{
								if (distance[i] == bestDistance)
								{
									neighbours[i] = t;
									distance[i] = thisDistance;
									break;
								}
							}
							bestDistance = distance[0];
							for (int i=1; i<neighbours.length; i++)
							{
								bestDistance = Math.max (bestDistance, distance[i]);
							}
							doWeedout = true;
						}
					}

					if (doWeedout && (found > 1))
					{
						// we are NOT interested in neighbours that are closer to another
						// neighbour than they are to us. 
						// when looking for 2 neighbours for "c" here: a--b--c-------e
						// we want b and e, not a and b, even if a is closer to c than e is.
						
						weedout = -1;
						for (int i=0; i<found; i++)
						{
							for (int j=i+1; j<found; j++)
							{
								double testdist = neighbours[i].distance (neighbours[j]);
								if (testdist < Math.min (distance[i], distance[j]))
								{
									weedout = (distance[i] < distance[j]) ? j : i;
/*
System.out.println ("weedout: mapping for "+thisPoint+" found="+found);
System.out.println ("   had neighbours "+neighbours[0].src +" and "+neighbours[1].src);
System.out.println ("   "+weedout+" must go");
*/
									break;
								}
							}
						}
						if (weedout > -1)
						{
							if (weedout < (found-1))
							{
								System.arraycopy (neighbours, weedout + 1, neighbours, weedout, found - weedout - 1);
								System.arraycopy (distance, weedout + 1, distance, weedout, found - weedout - 1);
							}
							neighbours[--found] = null;
						}
					}
				}
			}

			// System.out.println ("radius = "+radius + ", found "+found+" neighbours");
			if (found == neighbours.length) 
			{
				break;
			}
		}

		// only do something if we have the required number of neighbours.
		double xdiff = 0;
		double ydiff = 0;

		if (found > 1)
		{
			double totalDist = 1.0 / distance[0];
			for (int i=1; i < found; i++)
			{
				totalDist += 1.0 / distance[i];
			}

			for (int i=0; i < found; i++)
			{
				double factor = (1 / distance[i]) / totalDist;
				Point imgPoint = (Point) corrections.get (neighbours[i]);
				xdiff += (imgPoint.x - neighbours[i].x) * factor;
				ydiff += (imgPoint.y - neighbours[i].y) * factor;
			}
		}
		else if (found == 1)
		{
			Point imgPoint = (Point) corrections.get (neighbours[0]);
			double factor = 1 / Math.sqrt (Math.max (distance[0], 1.0));
			xdiff = (imgPoint.x - neighbours[0].x) * factor;
			ydiff = (imgPoint.y - neighbours[0].y) * factor;
// System.out.println ("sqrt dev = "+xdiff + ","+ydiff);
		}

		return (new double[] { xdiff, ydiff });
	}

	public void applyCorrection (Point thePoint)
	{
		double[] cor = computeCorrection (thePoint);
		thePoint.translate ((int) (cor[0] + 0.5), (int) (cor[1] + 0.5));
	}

	public static class Line
	{
		public Point p1;
		public Point p2;
	}

}
