/**
 * 
 * RoadHog Verifier
 * v0.9
 * (c) 2001 Frederik Ramm 
 * Released under GPL
 *
 * I know this needs documentation. 
 * See http://www.remote.org/frederik/roadhog/software/ for details.
 */

package org.remote.roadhog;

import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import gnu.getopt.Getopt;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.apache.xerces.parsers.DOMParser;

public class Verifier
{

	////////////////////////////////
	// @@ Properties
	////////////////////////////////

	private PrintStream reportingStream;		// where we log progress

	////////////////////////////////
	// @@ private data
	////////////////////////////////

	private HashMap journeys = new HashMap ();
	private HashMap vertices = new HashMap ();
	private HashMap edges = new HashMap ();
	private HashMap maps = new HashMap ();
	private DOMParser parser = new DOMParser ();

	////////////////////////////////
	// @@ Constructor
	////////////////////////////////

	public Verifier ()
	{

	}

	////////////////////////////////
	// @@ Public Methods
	////////////////////////////////

	public void readFile (String filename)
	throws ParseException, IOException, SAXException
	{
		XMLUtils.readFile (parser, filename, vertices, maps, edges, journeys, null, reportingStream, false, false);
	}

	public void verify ()
	throws IOException
	{
		report ("Verifying...");

		checkVertices ();
		checkEdges ();

		report ("Finished.");

	}

	////////////////////////////////
	// @@ Properties
	////////////////////////////////

	public void setReportingStream (PrintStream str)
	{
		reportingStream = str;
	}
	public PrintStream getReportingStream ()
	{
		return reportingStream;
	}


	////////////////////////////////
	// @@ Helpers
	////////////////////////////////

	private void report (String message)
	{
		if (reportingStream != null) reportingStream.println (message);
	}
	private void message (String message)
	{
		System.out.println (message);
	}

	private void checkVertices ()
	{
		for (Iterator i = vertices.values ().iterator (); i.hasNext (); )
		{
			Vertex v = (Vertex) i.next ();
			if (v.connector.length == 1)
			{
				report ("   vertex '" + v.id + "' is a dead end");
			}
			int count [] = new int[v.connector.length];
			for (Iterator j = edges.values ().iterator (); j.hasNext (); )
			{
				Edge e = (Edge) j.next ();
				for (int k = 0; k<2; k++)
				{
					if (e.vertex[k] == v)
					{
						for (int l = 0; l<v.connector.length; l++)
						{
							if (e.connector[k].equals (v.connector[l]))
							{
								count[l]++;
							}
						}
					}
				}
			}
			boolean isolated = true;
			for (int j = 0; j<v.connector.length; j++)
			{
				if (count[j] > 0) isolated = false;
				if (count[j] != 1)
				{
					StringBuffer s = new StringBuffer ("   vertex '");
					s.append (v.id);
					s.append ("': connector '");
					s.append (v.connector[j]);
					s.append ("' has ");
					if (count[j] == 0) s.append ("no"); else s.append (count[j]);
					s.append (" edges");
					message (s.toString ());
				}
			}
			if (isolated)
			{
				message ("   vertex '" + v.id + "' is ISOLATED");
			}
		}
	}


	private void checkEdges ()
	{
		for (Iterator i = edges.values ().iterator (); i.hasNext (); )
		{
			Edge e = (Edge) i.next ();
			HashMap out = new HashMap ();
			HashMap ret = new HashMap ();

			for (Iterator k = journeys.values ().iterator (); k.hasNext (); )
			{
				Journey j = (Journey) k.next ();
				if ((j.getAlternative () != null) && (!j.getAlternative ().equals (e.alternative))) continue;
				Double d = new Double (j.priority);
				if ((j.from == e.vertex[0]) && (j.to == e.vertex[1]))
				{
					if (out.get(d) != null)
					{
						out.put (d, new Integer (1 + ((Integer)out.get(d)).intValue ()));
					}
					else
					{
						out.put (d, new Integer (1));
					}
				}
				else if ((j.from == e.vertex[1]) && (j.to == e.vertex[0]))
				{
					if (ret.get(d) != null)
					{
						ret.put (d, new Integer (1 + ((Integer)ret.get(d)).intValue ()));
					}
					else
					{
						ret.put (d, new Integer (1));
					}
				}
			}

			if ((out.size () == 0) && (ret.size () == 0))
			{
				message ("   edge '" + e.getLongHandle () + "' has no journeys");
			}
			else if (out.size () == 0)
			{
				message ("   edge '" + e.getLongHandle () + "' has no journeys in " + e.vertex[0].id + "->" + e.vertex[1].id + " direction");
			}
			else if (ret.size () == 0)
			{
				message ("   edge '" + e.getLongHandle () + "' has no journeys in " + e.vertex[1].id + "->" + e.vertex[0].id + " direction");
			}
			for (Iterator l = out.keySet ().iterator (); l.hasNext (); )
			{
				Double key = (Double) l.next ();
				Integer val = (Integer) out.get (key);
				if (val.intValue () != 1)
				{
					message ("   edge '" + e.getLongHandle () + "' has "+val+" journeys of priority="+key+" in " + e.vertex[0].id + "->" + e.vertex[1].id + " direction");
				}
			}
			for (Iterator l = ret.keySet ().iterator (); l.hasNext (); )
			{
				Double key = (Double) l.next ();
				Integer val = (Integer) ret.get (key);
				if (val.intValue () != 1)
				{
					message ("   edge '" + e.getLongHandle () + "' has "+val+" journeys of priority="+key+" in " + e.vertex[1].id + "->" + e.vertex[0].id + " direction");
				}
			}

		}
	}


/*
	private void processEdge (Edge edge, String staticDir, String dynamicDir)
	throws IOException
	{
		String handle = edge.getHandle ();
		report ("  processing Edge: " + handle + "...");
		
		String staticEdgeDir = staticDir + handle + "/";
		String dynamicEdgeDir = dynamicDir + handle + "/";
		// HashMap journeyList = new HashMap ();

		// this processes all the photos
		for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
		{
			Journey j = (Journey) i.next ();
			if (j.edgeMatches (edge))
			{
				// journeyList.put (add, j.description);
				processJourney (j, edge, staticEdgeDir + j.getHandle () + "/", dynamicEdgeDir + j.getHandle () + "/");
			}
		}
		
		
	}

	public void createPlaceIndex (String filename)
	throws IOException
	{
		// this will be a hash map whose elements are hash sets of edges
		HashMap vias = new HashMap ();

		for (Iterator i = edges.values ().iterator (); i.hasNext (); )
		{
			Edge e = (Edge) i.next ();
			pushOntoViaMap (vias, e, e.vertex[0].name);
			pushOntoViaMap (vias, e, e.vertex[1].name);
			pushOntoViaMap (vias, e, e.road);
			if (e.via != null)
			{
				for (Enumeration v = e.via.elements(); v.hasMoreElements() ;) 
				{
					pushOntoViaMap (vias, e, (String) v.nextElement ());
				}
			}
		}

		// write list of all journeys to the data file
		PrintWriter p = new PrintWriter (new FileOutputStream (filename));
		int index = 0;
		HashMap journeyIndex = new HashMap ();

		for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
		{
			Journey j = (Journey) i.next ();
			Edge e = j.findEdge (edges.values ().iterator ());
			if (e != null)
			{
				journeyIndex.put (j, new Integer (index++));
				p.print (e.getHandle ());
				p.print ("/");
				p.print (j.getHandle ());
				p.print (":");
				p.print ("Journey from "+j.from.name+" to "+j.to.getUltimateDestination (e));
				if (j.getDetail () != null)
				{
					p.print (" (");
					p.print (j.getDetail ());
					p.print (")");
				}
				p.println ();
			}
		}
		p.println ();
	
		for (Iterator i = vias.keySet ().iterator (); i.hasNext (); )
		{
			String name = (String) i.next ();
			HashSet edgeSet = (HashSet) vias.get (name);
			String qualifier = "";
			int paren = name.indexOf ('(');
			boolean first = true;
			if (paren > -1)
			{
				qualifier = name.substring (paren + 1, name.length () - 1);
				name = name.substring (0, paren - 1);
			}
			p.print (name);
			p.print (":");
			p.print (qualifier);
			p.print (":");
			
			// now collect all journey indexes for this keyword
			// (all journeys that use an edge which claims this keyword)
			for (Iterator j = edgeSet.iterator (); j.hasNext (); )
			{
				Edge testEdge = (Edge) j.next ();
				for (Iterator k = journeys.values ().iterator (); k.hasNext (); )
				{
					Journey testJourney = (Journey) k.next ();
					if (testJourney.edgeMatches (testEdge))
					{
						if (!first) p.print (",");
						first = false;
						p.print (journeyIndex.get (testJourney));
					}
				}
			}
			p.println ();
		}
		p.close ();

	}

	public void createLocationIndex (String filename)
	throws IOException
	{
		Vector points = new Vector ();
		int index = 0;
		PrintWriter p = new PrintWriter (new FileOutputStream (filename));
		double minlong = Double.MAX_VALUE;
		double minlat = Double.MAX_VALUE;
		double maxlat = -Double.MAX_VALUE;

		// build index of all coordinate points
		for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
		{
			Journey j = (Journey) i.next ();
			Edge e = j.findEdge (edges.values ().iterator ());
			if (e != null)
			{
				p.print (e.getHandle ());
				p.print ("/");
				p.print (j.getHandle ());
				p.print (":");
				p.print ("Journey from "+j.from.name+" to "+ j.to.getUltimateDestination (e));
				if (j.getDetail() != null)
				{
					p.print (" (");
					p.print (j.getDetail ());
					p.print (")");
				}
				p.println ();
				Integer indexObject = new Integer (index++);
				for (int k = 0; k < j.photo.length; k++)
				{
					Waypoint wp = j.photo[k].waypoint;
					points.add (new Object[] { wp.coord, indexObject });
					if (wp.coord.x < minlong) minlong = wp.coord.x;
					if (wp.coord.y < minlat) minlat = wp.coord.y;
					if (wp.coord.y > maxlat) maxlat = wp.coord.y;
				}
			}
		}

		double latFactor = 111120;
		double medianLatitude = (maxlat + minlat) / 2;
		double longFactor = latFactor * Math.cos (Math.toRadians (medianLatitude));

		p.println ();
		p.print (minlong);
		p.print (":");
		p.print (maxlat);
		p.print (":");
		p.print (longFactor);
		p.print (":");
		p.println (latFactor);


		// sort
		Collections.sort (points, new Comparator () {
			public int compare (Object a, Object b) {
				double y1 = ((Point2D.Double) ((Object[])a)[0]).y;
				double y2 = ((Point2D.Double) ((Object[])b)[0]).y;
				return (y1>y2) ? -1 : (y1<y2) ? 1 : 0;
			}
		});

		// write
		for (Enumeration e = points.elements (); e.hasMoreElements (); )
		{
			Object[] array = (Object[]) e.nextElement ();
			Point2D.Double pt = (Point2D.Double) array[0];
			p.print ((int)((pt.x - minlong) * longFactor));
			p.print (":");
			p.print ((int)((maxlat - pt.y) * latFactor));
			p.print (":");
			double os[] = GridConverter.degreeToGrid (pt.y, pt.x);
			p.print ((int) (os[0] + 0.5));
			p.print (":");
			p.print ((int) (os[1] + 0.5));
			p.print (":");
			 
			p.println ((Integer) array[1]);
		}
		p.close ();

	}

	public void processJourney (Journey j, Edge edge, String staticDir, String dynamicDir)
	throws IOException
	{

		NumberFormat nf = NumberFormat.getInstance ();
		nf.setMinimumIntegerDigits ((j.photo.length < 100) ? 2 : 3);
		nf.setParseIntegerOnly (true);
		HashMap writers = new HashMap ();
		AffineTransform transform = new AffineTransform ();
		StringBuffer mapSizeLine = new StringBuffer ();
		StringBuffer imageSizeLine = new StringBuffer ();
		
		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			if (spec.mapScale != 0)
			{
				if (mapSizeLine.length () > 0) mapSizeLine.append (':');
				mapSizeLine.append (spec.name);
			}
			if (spec.imageX != 0)
			{
				if (imageSizeLine.length () > 0) imageSizeLine.append (':');
				imageSizeLine.append (spec.name);
			}
		}

		// create directories if required; initialize the journey control files
		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			File path = new File (staticDir + spec.name);
			path.mkdirs ();

			if (spec.mapScale == 0) continue;

			path = new File (dynamicDir + spec.name);
			path.mkdirs ();
			report ("    writing " + dynamicDir + spec.name + "/control.txt");
			PrintWriter p = new PrintWriter (new FileOutputStream (dynamicDir +
				spec.name + "/control.txt"));
			writers.put (spec.name, p);

			// generate a line that reads like this:
			// Cove:Poolewe:Wester Ross (Scotland):B8057:64
			p.print (j.from.name);
			p.print (":");
			p.print (j.to.getUltimateDestination (j.getEdge ()));
			p.print ("::"); // area code left empty
			if (edge.road != null) p.print (edge.road);
			p.print (":");
			p.println (j.photo.length);

			// generate a line like this:
			// 256:192:bla-blub:3:2:116:163:54,54,55:65,51
			// decodes as imagewidth:imageheight:mapname:maprows:mapcols:mapwidth:mapheight:
			//       rowheights:colwidths (for map segments; comma separated)
			p.print (spec.imageX);
			p.print (":");
			p.print (spec.imageY);
			p.print (":");
			p.print (edge.map.getHandle ());
			p.print (":");
			if (spec.mapScale == 0)
			{
				p.print ("0:0:0:0::");
			}
			else
			{
				MapInfo mi = (MapInfo) edge.map.mapInfo.get (spec.name);
				p.print (mi.width);
				p.print (":");
				p.print (mi.height);
				p.print (":");
				p.print (mi.rows);
				p.print (":");
				p.print (mi.columns);
				p.print (":");

				if ((mi.rows == 1) && (mi.columns == 1))
				{
					p.print (":");
				}
				else
				{
					for (int i=0; i<mi.rows; i++)
					{
						p.print (mi.partRect[i][0].height);
						if (i<mi.rows-1) p.print (",");
					}
					p.print (":");
					for (int i=0; i<mi.columns; i++)
					{
						p.print (mi.partRect[0][i].width);
						if (i<mi.columns-1) p.print (",");
					}
				}
			}

			p.println ();	

			// generate a line about the available sizes in ascending order like this:
			// tiny:small:medium:large:huge
			p.println (mapSizeLine);	
			p.println (imageSizeLine);	
			p.println ();
		}

		for (int i=0; i < j.photo.length; i++)
		{
			// this produces one jpg file for each photo and size class
			String fileName = nf.format (i+1) + ".jpg";

			// convert this photo's GPS coordinates into pixels 
			j.photo[i].waypoint.determineMapPixel (edge.map);
			Point pixel = j.photo[i].waypoint.mapPixel;

			MagickImage srcImage = null;

			if (!mapsOnly)
			{
				if (incremental)
				{
					// we may not have to load the image at all - only load if at least
					// one of the target images diesn't exist
					for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
					{
						String name = staticDir + ((SizeSpec) e.nextElement ()).name + "/" + fileName;
						if (! fileExists (name))
						{
							srcImage = fetchAndPreprocessImage (j.photo[i]);
							break;
						}
					}
				}
				else
				{
					srcImage = fetchAndPreprocessImage (j.photo[i]);
				}
			}

			for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
			{
				SizeSpec spec = (SizeSpec) e.nextElement ();

				if (!mapsOnly)
				{
					String name = staticDir + spec.name + "/" + fileName;
					if (!(incremental && fileExists (name)))
					{
						report ("      writing " + name);
						try
						{
							MagickImage newImage = srcImage.scaleImage (spec.imageX, spec.imageY);
							ImageInfo newImageInfo = new ImageInfo (name);
							newImage.setFileName (name);
							newImageInfo.setQuality (spec.imageQuality);
							newImage.writeImage (newImageInfo);
							newImage.destroyImage ();
							increaseMagickUsage ();
						}
						catch (MagickException m)
						{
							throw new IOException (m.getMessage ());
						}
					}
				}

				if (spec.mapScale == 0) continue;

				MapInfo mi = (MapInfo) edge.map.mapInfo.get (spec.name);
				Point scaledPos = new Point (pixel.x * spec.mapScale / 100, 
					pixel.y * spec.mapScale / 100);
				int radius = (int) (8.0 * Math.sqrt ((double) spec.mapScale / 100.0));
				Rectangle pointEnclosingRect = new Rectangle (scaledPos.x - radius, scaledPos.y - radius,
					2*radius, 2*radius);
				PrintWriter pw = (PrintWriter) writers.get (spec.name);
				StringBuffer mapSegName = new StringBuffer ();

				// find related map segment 
				// boolean preferred = true;
				for (int row=0; row<mi.rows; row++)
				for (int col=0; col<mi.columns; col++)
				{
					if (mi.partRect[row][col].intersects (pointEnclosingRect))
					{
						mi.dotImageRequests[row][col].put (scaledPos, null);
						if (mapSegName.length () != 0) mapSegName.append (",");
						mapSegName.append (row+1);
						mapSegName.append ("x");
						mapSegName.append (col+1);
					}
				}

				// stuff info about this image into control file
				pw.print (fileName);
				pw.print (":");
				pw.print (scaledPos.x);
				pw.print (":");
				pw.print (scaledPos.y);
				pw.print (":");
				pw.print (mapSegName);
				pw.print (":");
				pw.print (j.photo[i].waypoint.prettyLatitude ());
				pw.print (":");
				pw.print (j.photo[i].waypoint.prettyLongitude ());
				pw.print (":");
				pw.print ((int) (j.photo[i].waypoint.time / 1000));
				pw.print (":");
				pw.print (j.photo[i].waypoint.speed);
				pw.print (":");
				pw.println (j.photo[i].waypoint.heading);
				
			}

			if (srcImage != null) srcImage.destroyImage ();
		}

		// where we are
		Vertex ici = j.to;
		String leg = edge.connector[(edge.vertex[0] == ici) ? 0 : 1];
		int fromConnector = ici.findConnectorIndex (leg);
		if (fromConnector == -1)
		{
			throw new IOException ("Edge points to non-existent connector '"+leg+"' at vertex '"+ici.name+"'");
		}

		// finish off the print writers by putting an empty line and information about
		// all connecting journeys at the end.

		for (Enumeration e = sizeSpecs.elements(); e.hasMoreElements() ;) 
		{
			SizeSpec spec = (SizeSpec) e.nextElement ();
			if (spec.mapScale == 0) continue;

			PrintWriter pw = (PrintWriter) writers.get (spec.name);
			
			pw.println ();


			pw.println (ici.detail.get("en")); // FIXME write all languages
			
			// one line for each connection.
			// format like this:
			// o:270:A832:Aultbea:poolewe-aultbea:out:april
			// type (o=option d=default b=back):turn (degrees):road name:destination:edge handle:
			// journey id

			for (Iterator i = journeys.values ().iterator (); i.hasNext (); )
			{
				Journey jj = (Journey) i.next ();
				if (jj.from != ici) continue;
				Edge ee = jj.findEdge (edges.values ().iterator ());
				String theConnector = ee.connector [(jj.from == ee.vertex[0]) ? 0 : 1];

				int onDir = -1;
				int onConnector;

				onConnector = ici.findConnectorIndex (theConnector);
				if (onConnector == -1)
				{
					throw new IOException ("Edge points to non-existent connector '" +
						theConnector + "' at vertex '"+jj.from.name+"'");
				}
				onDir = ici.direction[onConnector] - ici.direction[fromConnector] + 540;
				if (onDir > 360) onDir -= 360;
				if (onDir > 360) onDir -= 360;
				pw.print (ici.isPassthrough (fromConnector, onConnector) ? "d" : (fromConnector == onConnector) ? "b" : "o");
				pw.print (":");
				pw.print (jj.priority);
				pw.print (":");
				if (jj.getDetail () != null) pw.print (jj.getDetail ());
				pw.print (":");
				pw.print (onDir);
				pw.print (":");
				pw.print (ee.road);
				pw.print (":");
				pw.print (":");
				pw.print (ee.getHandle ());
				pw.print (":");
				pw.println (jj.getHandle ());
			}

			pw.close ();
		}
		
	}

*/
	////////////////////////////////
	// @@ Static "main" method
	////////////////////////////////

	/**
	 * Main method. 
	 * Parses arguments and calls appropiate methods on verifier object;
	 * finally asks verifier object to proceed.
	 */
	public static void main(String argv[])
	{

		Verifier theVerifier = new Verifier ();
		Getopt g = new Getopt ("verifier", argv, "hv");
		int c;

		g.setOpterr (false);

		while ((c = g.getopt ()) != -1)
		{
			switch (c)
			{
			  case 'v':
				theVerifier.setReportingStream (System.out);
				break;
			  default:
				if (c == ':') 
				{
					System.out.println ("-" + (char)g.getOptopt () + " requires an argument\n");
				}
				else if (c != 'h')
				{
					System.out.println ("Invalid option: -" + (char) g.getOptopt () + "\n");
				}
				System.out.println ("Usage:");
				System.out.print ("java " + theVerifier.getClass ().getName ());
				System.out.println (" [option]... [inputfile]...");
				System.out.println ("   -h             displays this help");
				System.out.println ("   -v             verbose progress report");
				System.out.println ("   -q             quick operation - no brightness/contrast oeprations");
				System.out.println ("   -d path        output destination path");
				System.out.println ("   -m             map processing only");
				System.out.println ("   -f             re-create files even if they exist");
				return;
			}
		}

		for (int i = g.getOptind(); i < argv.length; i++)
		{
			try
			{
				theVerifier.readFile (argv[i]);
			}
			catch (Exception ex)
			{
				System.err.println ("Error while processing file '" + argv[i] + "':");
				System.err.println ("   " + ex.getClass () + ": " + ex.getMessage ());
				if (theVerifier.getReportingStream() != null) ex.printStackTrace ();
			}
		}

		try
		{
			theVerifier.verify ();
		}
		catch (Exception ex)
		{
			System.err.println ("Error during verification:");
			System.err.println ("   " + ex.getClass () + ": " + ex.getMessage ());
			if (theVerifier.getReportingStream() != null) ex.printStackTrace ();
		}

		System.exit (0);
	}

}
